#
# pyFlow - a lightweight parallel task engine
#
# Copyright (c) 2012-2015 Illumina, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
# WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#


"""
bwaworkflow -- a pyflow demonstration module

This is a quick-and-dirty BCL to BWA BAM workflow to demonstrate
how pyflow could be used on a production-scale problem.

__author__ = "Christopher Saunders"
"""


import os.path
import sys

# In production, pyflow can either be installed, or we can distribute
# workflow to external users with pyflow in the same directory/fixed
# relative directory or a configured directory macro-ed in by cmake,
# etc
#
# For now we add the module path by hand:
#
scriptDir = os.path.abspath(os.path.dirname(__file__))
sys.path.append(scriptDir + "/../../src")

from pyflow import WorkflowRunner


#
# utility methods:
#

def ensureDir(d):
    """
    make directory if it doesn't already exist, raise exception is something else is in the way:
    """
    if os.path.exists(d):
        if not os.path.isdir(d) :
            raise Exception("Can't create directory: %s" % (d))
    else :
        os.makedirs(d)


def skipJoin(sep, a, b) :
    if a == "" : return b
    elif b == "" : return a
    return a + sep + b


def preJoin(a, b) :
    return skipJoin('_', a, b)


#
# All of these "flow" functions take a set of task dependencies as
# input and report a set of tasks on output, and thus are designed to
# be plugged together to create workflows which are initiated in
# the WorkflowRunner.workflow() method.
#
# Note that this style is not a design constraint of pyflow, it was
# just one natural way to write the bwa workflow, and demonstrates an
# extensible model wherein various flow functions could be stored in
# external modules and combined as required.
#
# Note that these flow functions are written to lookup dependencies
# from the WorkflowRunner class, so they are really class
# methods. Although they could also lookup data from the derived BWA
# class, they don't -- this allows them to be reused by other
# WorkflowRunner classes.
#



def casava18BclToFastqFlow(self, taskPrefix="", dependencies=set()) :
    """
    CASAVA 1.8 bcl to fastq conversion

    This assumes the bclBasecallsDir is generated in a CASAVA 1.8
    compatible format, and uses CASAVA 1.8 to convert to fastq

    This demonstrates pyflow's makefile handling option, where
    you specify a makefile directory instead of a regular command, and
    pyflow runs make/qmake according to the run mode.

    params:
    casavaDir
    bclBasecallsDir
    flowcellFastqDir
    bclTilePattern
    bclToFastqMaxCores
    """

    # configure bcl2fastq makefile:
    configBclToFastqCmd = "perl %s/bin/configureBclToFastq.pl" % (self.params.casavaDir)
    configBclToFastqCmd += " --input-dir=%s" % self.params.bclBasecallsDir
    configBclToFastqCmd += " --output-dir=%s" % self.params.flowcellFastqDir
    configBclToFastqCmd += " --force"  # always a good idea for CASAVA
    configBclToFastqCmd += " --ignore-missing-bcl"
    configBclToFastqCmd += " --ignore-missing-stats"
    configBclToFastqCmd += " --ignore-missing-control"
    if self.params.bclTilePattern != None :
        configBclToFastqCmd += " --tiles=%s" % (self.params.bclTilePattern)

    # run configuration:
    configLabel = self.addTask(preJoin(taskPrefix, "configBclToFastq"), configBclToFastqCmd, isForceLocal=True, dependencies=dependencies)

    # for the bcl to fastq step, we use another workflow manager, so
    # we just run it as one huge task and handle the mode ourselves:
    nCores = self.getNCores()
    mode = self.getRunMode()

    maxCores = self.params.bclToFastqMaxCores
    if (nCores == "unlimited") or (nCores > maxCores) :
        nCores = maxCores

    # run the fastq conversion:
    bclToFastqLabel = self.addTask(preJoin(taskPrefix, "bclToFastq"),
                                 self.params.flowcellFastqDir,
                                 nCores=nCores,
                                 dependencies=configLabel,
                                 isCommandMakePath=True)

    return set([bclToFastqLabel])



class FastqPairToBwaBamFlow(WorkflowRunner) :
    """
    Given a read1 and read2 pair of fastq files, create an aligned and
    sorted bamFile.  Don't delete input fastaq files.
    """

    def __init__(self, params, suggestedAlignThreadCount=2) :
        """
        suggestedAlignThreadCount -- Number of threads to use in bwa aln
                                     step. The workflow will lower this if
                                     it exceeds the total number of cores
                                     available in the run, or if it
                                     exceeds alnMaxCores

        params:
        fastq1File
        fastq2File
        bamFile
        alnMaxCores
        bwaBin
        genomeFasta
        samtoolsBin
        samtoolsSortMemPerCore
        isKeepFastq
        """
        self.params = params
        self.suggestedAlignThreadCount = suggestedAlignThreadCount


    def workflow(self) :
        bamDir = os.path.dirname(self.params.bamFile)
        ensureDir(bamDir)

        (bamPrefix, bamExt) = os.path.splitext(self.params.bamFile)

        # must end in ".bam" for samtools
        if bamExt != ".bam" :
            raise Exception("bamFile argument must end in '.bam'. bamFile is: %s" % (bamFile))
        if bamPrefix == "" :
            raise Exception("bamFile argument must have a prefix before the '.bam' extension.")

        # assuming many fastq pairs are running, good total throughput given cluster nodes with 2G of ram each
        # should be achieved by given the align processes 2 threads each:

        # grab total cores to make sure we don't exceed it:
        totalCores = self.getNCores()

        #
        # setup aln step:
        #

        # set alnCores
        alnCores = int(self.suggestedAlignThreadCount)
        if (totalCores != "unlimited") and (alnCores > totalCores) :
            alnCores = int(totalCores)
        if (alnCores > self.params.alnMaxCores) :
            alnCores = int(self.params.alnMaxCores)

        bwaBaseCmd = "%s aln -t %i %s" % (self.params.bwaBin, alnCores, self.params.genomeFasta)

        peDependencies = set()

        def getReadLabel(i) : return "Read%iBwaAlign" % (i)
        def getReadSaiFile(i) : return "%s.read%i.sai" % (self.params.bamFile, i)
        def getReadFastqFile(i) : return (self.params.fastq1File, self.params.fastq2File)[i - 1]

        for read in (1, 2) :
            readAlnCmd = "%s %s >| %s" % (bwaBaseCmd, getReadFastqFile(read), getReadSaiFile(read))
            peDependencies.add(self.addTask(getReadLabel(read), readAlnCmd, nCores=alnCores))

        #
        # setup sampe step:
        #

        # with all the pipes, the sampe step is probably a 2 core? this lets sort use more mem too:
        peCores = 2
        if (totalCores != "unlimited") and (peCores > totalCores) :
            peCores = int(totalCores)

        peCmd = "%s sampe %s %s %s %s %s" % (self.params.bwaBin, self.params.genomeFasta,
                                            getReadSaiFile(1), getReadSaiFile(2),
                                            getReadFastqFile(1), getReadFastqFile(2))

        peCmd += " | %s view -uS -" % (self.params.samtoolsBin)

        # For a real pipeline, we'd probably prefer Picard sort, but I don't want to add another
        # dependency to the trial workflow:
        #
        peCmd += " | %s sort -m %i - %s" % (self.params.samtoolsBin,
                                          self.params.samtoolsSortMemPerCore,  # *peCores, need to leave memory for bwa...
                                          bamPrefix)

        peTaskLabel = self.addTask("BwaSamPESort", peCmd, nCores=peCores, dependencies=peDependencies)

        # delete sai files:
        rmCmd = "rm -f"
        for read in (1, 2) :
            rmCmd += " %s" % (getReadSaiFile(read))
        self.addTask("RmSai", rmCmd, dependencies=peTaskLabel, isForceLocal=True)


        # optionally delete input fastqs:
        if not self.params.isKeepFastq :
            fastqRmCmd = "rm -f"
            for read in (1, 2) :
                fastqRmCmd += " %s" % (getReadFastqFile(read))
            self.addTask("RmFastq", fastqRmCmd, dependencies=peTaskLabel, isForceLocal=True)




class FileDigger(object) :
    """
    Digs into a well-defined directory structure with prefixed
    folder names to extract all files associated with
    combinations of directory names.

    This is written primarily to go through the CASAVA 1.8 output
    structure.

    #casava 1.8 fastq example:
    fqDigger=FileDigger('.fastq.gz',['Project_','Sample_'])
    """

    def __init__(self, targetExtension, prefixList) :
        self.targetExtension = targetExtension
        self.prefixList = prefixList

    def getNextFile(self, dir, depth=0, ans=tuple()) :
        """
        generator of a tuple: (flowcell,project,sample,bamfile)
        given a multi-flowcell directory
        """
        if depth < len(self.prefixList) :
            for d in os.listdir(dir) :
                nextDir = os.path.join(dir, d)
                if not os.path.isdir(nextDir) : continue
                if not d.startswith(self.prefixList[depth]) : continue
                value = d[len(self.prefixList[depth]):]
                for val in self.getNextFile(nextDir, depth + 1, ans + tuple([value])) :
                    yield val
        else:
            for f in os.listdir(dir) :
                file = os.path.join(dir, f)
                if not os.path.isfile(file) : continue
                if not f.endswith(self.targetExtension) : continue
                yield ans + tuple([file])



def flowcellDirFastqToBwaBamFlow(self, taskPrefix="", dependencies=set()) :
    """
    Takes as input 'flowcellFastqDir' pointing to the CASAVA 1.8 flowcell
    project/sample fastq directory structure. For each project/sample,
    the fastqs are aligned using BWA, sorted and merged into a single
    BAM file. The bam output is placed in a parallel project/sample
    directory structure below 'flowcellBamDir'

    params:
    samtoolsBin
    flowcellFastqDir
    flowcellBamDir

    calls:
    FastqPairToBwaBamFlow
        supplies:
        bamFile
        fastq1File
        fastq2File
    """

    #
    # 1. separate fastqs into matching pairs:
    #
    fqs = {}
    fqDigger = FileDigger(".fastq.gz", ["Project_", "Sample_"])
    for (project, sample, fqPath) in fqDigger.getNextFile(self.params.flowcellFastqDir) :
        if (self.params.sampleNameList != None) and \
           (len(self.params.sampleNameList) != 0) and \
           (sample not in self.params.sampleNameList) : continue

        fqFile = os.path.basename(fqPath)
        w = (fqFile.split(".")[0]).split("_")
        if len(w) != 5 :
            raise Exception("Unexpected fastq filename format: '%s'" % (fqPath))

        (sample2, index, lane, read, num) = w
        if sample != sample2 :
            raise Exception("Fastq name sample disagrees with directory sample: '%s;" % (fqPath))

        key = (project, sample, index, lane, num)
        if key not in fqs : fqs[key] = [None, None]

        readNo = int(read[1])
        if fqs[key][readNo - 1] != None :
            raise Exceptoin("Unresolvable repeated fastq file pattern in sample: '%s'" % (fqPath))
        fqs[key][readNo - 1] = fqPath


    ensureDir(self.params.flowcellBamDir)

    #
    # 2. run all fastq pairs through BWA:
    #
    nextWait = set()

    for key in fqs.keys() :
        (project, sample, index, lane, num) = key
        sampleBamDir = os.path.join(self.params.flowcellBamDir, "Project_" + project, "Sample_" + sample)
        ensureDir(sampleBamDir)
        keytag = "_".join(key)
        self.params.bamFile = os.path.join(sampleBamDir, keytag + ".bam")
        self.params.fastq1File = fqs[key][0]
        self.params.fastq2File = fqs[key][1]
        nextWait.add(self.addWorkflowTask(preJoin(taskPrefix, keytag), FastqPairToBwaBamFlow(self.params), dependencies=dependencies))

    return nextWait



class FlowcellDirFastqToBwaBamFlow(WorkflowRunner) :
    """
    Takes as input 'flowcellFastqDir' pointing to the CASAVA 1.8 flowcell
    project/sample fastq directory structure. For each project/sample,
    the fastqs are aligned using BWA, sorted and merged into a single
    BAM file. The bam output is placed in a parallel project/sample
    directory structure below 'flowcellBamDir'

    params:
    flowcellFastqDir
    flowcellBamDir
    """

    def __init__(self, params) :
        self.params = params

    def workflow(self) :
        flowcellDirFastqToBwaBamFlow(self)





# use a really boring flowcell label everywhere right now:
def getFlowcellLabel(self, i) :
    return "Flowcell_FC%i" % (i)




def casava18BclToBamListFlow(self, taskPrefix="", dependencies=set()) :
    """
    Runs bcl conversion and alignment on multiple flowcells for a subset of samples.
    Writes BAM files to parallel fastq Project/Sample directory structure. Does not
    merge individual BAMs. Deletes fastqs on alignment when option is set to do so.

    params:
    allFlowcellDir
    bclBasecallsDirList
    bclTilePatternList

    calls:
    casava18BclToFastqFlow
        supplies:
        bclBasecallsDir
        flowcellFastqDir
    FlowcellDirFastqToBwaBamFlow
        supplies:
        flowcellFastqDir
        flowcellBamDir

    """

    ensureDir(self.params.allFlowcellDir)

    # first bcl->fastq->bwa bam for requested samples in all flowcells:
    nextWait = set()
    for i, self.params.bclBasecallsDir in enumerate(self.params.bclBasecallsDirList) :
        flowcellLabel = getFlowcellLabel(self, i)
        flowcellDir = os.path.join(self.params.allFlowcellDir, flowcellLabel)

        ensureDir(flowcellDir)

        self.params.flowcellFastqDir = os.path.join(flowcellDir, "fastq")
        self.params.flowcellBamDir = os.path.join(flowcellDir, "bam")
        if self.params.bclTilePatternList == None :
            self.params.bclTilePattern = None
        else :
            self.params.bclTilePattern = self.params.bclTilePatternList[i]

        fastqFinal = casava18BclToFastqFlow(self, taskPrefix=flowcellLabel)

        label = preJoin(taskPrefix, "_".join((flowcellLabel, "FastqToBwaBam")))
        nextWait.add(self.addWorkflowTask(label, FlowcellDirFastqToBwaBamFlow(self.params), dependencies=fastqFinal))

    return nextWait




def mergeBamListFlow(self, taskPrefix="", dependencies=set()) :
    """
    Take a list of sorted bam files from the same sample, merge them together,
    and delete input bams, final output to mergeBamName

    params:
    mergeBamList
    mergeBamName
    samtoolsBin
    """

    for bamFile in self.params.mergeBamList :
        if not os.path.isfile(bamFile) :
            raise Exception("Can't find bam file: '%s'" % (bamFile))

    mergeTasks = set()
    mergeLabel = preJoin(taskPrefix, "merge")
    if len(self.params.mergeBamList) > 1 :
        mergeCmd = "%s merge -f %s %s" % (self.params.samtoolsBin, self.params.mergeBamName, " ".join(self.params.mergeBamList))
        mergeTasks.add(self.addTask(mergeLabel, mergeCmd, dependencies=dependencies, isTaskStable=False))

        rmCmd = "rm -f"
        for bamFile in self.params.mergeBamList :
            rmCmd += " %s" % (bamFile)

        self.addTask(preJoin(taskPrefix, "rmBam"), rmCmd, dependencies=mergeLabel, isForceLocal=True)
    elif len(self.params.mergeBamList) == 1 :
        mvCmd = "mv %s %s" % (self.params.mergeBamList[0], self.params.mergeBamName)
        # *must* have same taskLabel as merge command for continuation
        # to work correctly because of the potential for partial
        # deletion of the input bam files:
        mergeTasks.add(self.addTask(mergeLabel, mvCmd, dependencies=dependencies, isForceLocal=True, isTaskStable=False))

    return mergeTasks



def flowcellBamListMergeFlow(self, taskPrefix="", dependencies=set()) :
    """
    given a root flowcell directory and list of samples, merge sample
    bams across flowcells and dedup.

    ?? Will we be in a situation where sample has more than one library
    -- this affects the debup order & logic ??

    params:
    allFlowcellDir
    mergedDir
    sampleNameList
    picardDir

    calls:
    mergeBamListFlow
        supplies:
        mergeBamList
        mergeBamName
    """

    #
    # 1) get a list of bams associated with each project/sample combination:
    #

    # TODO: what if there's an NFS delay updating all the bams while
    # we're reading them out here? make this process more robust -- we
    # should know how many BAM's we're expecting, in a way that's
    # robust to interuption/restart
    #
    bams = {}
    bamDigger = FileDigger(".bam", ["Flowcell_", "bam", "Project_", "Sample_"])
    for (flowcell, nothing, project, sample, bamFile) in bamDigger.getNextFile(self.params.allFlowcellDir) :
        if (self.params.sampleNameList != None) and \
           (len(self.params.sampleNameList) != 0) and \
           (sample not in self.params.sampleNameList) : continue
        key = (project, sample)
        if key not in bams : bams[key] = []
        bams[key].append(bamFile)

    mergedBamExt = ".merged.bam"
    markDupBamExt = ".markdup.bam"

    #
    # 2) merge and delete smaller bams:
    #
    mergedBams = {}

    mergedBamDir = os.path.join(self.params.mergedDir, "bam")
    sampleTasks = {}
    if len(bams) :  # skip this section if smaller bams have already been deleted
        ensureDir(mergedBamDir)

        for key in bams.keys() :
            (project, sample) = key
            mergedSampleDir = os.path.join(mergedBamDir, "Project_" + project, "Sample_" + sample)
            ensureDir(mergedSampleDir)
            self.params.mergeBamList = bams[key]
            self.params.mergeBamName = os.path.join(mergedSampleDir, sample + mergedBamExt)
            mergedBams[key] = self.params.mergeBamName
            outTaskPrefix = preJoin(taskPrefix, "_".join(key))
            sampleTasks[key] = mergeBamListFlow(self, outTaskPrefix, dependencies)

    if not os.path.isdir(mergedBamDir) : return


    #
    # 3) mark dup:
    #

    # mergedBams contains all bams from the current run, we also add any from a
    # previous interupted run:
    mergedBamDigger = FileDigger(mergedBamExt, ["Project_", "Sample_"])
    for (project, sample, bamFile) in mergedBamDigger.getNextFile(mergedBamDir) :
        key = (project, sample)
        if key in mergedBams :
            assert (mergedBams[key] == bamFile)
        else :
            mergedBams[key] = bamFile

    nextWait = set()
    totalCores = self.getNCores()

    for sampleKey in mergedBams.keys() :
        markDupDep = set()
        if sampleKey in sampleTasks : markDupDep = sampleTasks[sampleKey]

        fullName = "_".join(sampleKey)

        markDupBamFile = mergedBams[sampleKey][:-(len(mergedBamExt))] + markDupBamExt
        markDupMetricsFile = markDupBamFile[:-(len(".bam"))] + ".metrics.txt"
        markDupTmpDir = markDupBamFile + ".tmpdir"

        # for now, solve the memory problem with lots of threads:
        nCores = 4
        if (totalCores != "unlimited") and (totalCores < nCores) :
            nCores = totalCores
        gigs = 2 * nCores
        javaOpts = "-Xmx%ig" % (gigs)
        markDupFiles = "INPUT=%s OUTPUT=%s METRICS_FILE=%s" % (mergedBams[sampleKey], markDupBamFile, markDupMetricsFile)
        markDupOpts = "REMOVE_DUPLICATES=false ASSUME_SORTED=true VALIDATION_STRINGENCY=SILENT CREATE_INDEX=true TMP_DIR=%s" % (markDupTmpDir)
        markDupJar = os.path.join(self.params.picardDir, "MarkDuplicates.jar")
        markDupCmd = "java %s -jar %s %s %s" % (javaOpts, markDupJar, markDupFiles, markDupOpts)
        markDupTask = self.addTask(preJoin(taskPrefix, fullName + "_dupmark"), markDupCmd, dependencies=markDupDep)

        # link index filename to something samtools can understand:
        #
        markDupPicardBaiFile = markDupBamFile[:-(len(".bam"))] + ".bai"
        markDupSamtoolsBaiFile = markDupBamFile + ".bai"
        indexLinkCmd = "ln %s %s" % (markDupPicardBaiFile, markDupSamtoolsBaiFile)
        indexLinkTask = self.addTask(preJoin(taskPrefix, fullName + "_indexLink"), indexLinkCmd, dependencies=markDupTask, isForceLocal=True)

        nextWait.add(indexLinkTask)

        # delete TmpDir:
        #
        rmMarkDupTmpCmd = "rm -rf %s" % (markDupTmpDir)
        self.addTask(preJoin(taskPrefix, fullName + "_rmMarkDupTmp"), rmMarkDupTmpCmd, dependencies=markDupTask, isForceLocal=True)

        # now remove the original file:
        #
        rmCmd = "rm -f %s" % (mergedBams[sampleKey])
        self.addTask(preJoin(taskPrefix, fullName + "_rmMerge"), rmCmd, dependencies=markDupTask, isForceLocal=True)

    return nextWait




class FlowcellBamListMergeFlow(WorkflowRunner) :

    def __init__(self, params) :
        self.params = params

    def workflow(self) :
        flowcellBamListMergeFlow(self)



class BWAWorkflow(WorkflowRunner) :
    """
    pyflow BCL to BAM BWA workflow
    """

    def __init__(self, params) :
        self.params = params

        # make sure working directory is setup:
        self.params.outputDir = os.path.abspath(self.params.outputDir)
        ensureDir(self.params.outputDir)

        self.params.allFlowcellDir = os.path.join(self.params.outputDir, "flowcell_results")
        self.params.mergedDir = os.path.join(self.params.outputDir, "merged_results")

        # Verify/manipulate various input options:
        #
        # this is mostly repeated in the conflig script now... get this minimized with auto verification:
        #
        self.params.bclBasecallsDirList = map(os.path.abspath, self.params.bclBasecallsDirList)
        for dir in self.params.bclBasecallsDirList :
            if not os.path.isdir(dir) :
                raise Exception("Input BCL basecalls directory not found: '%s'" % (dir))

        self.params.samtoolsSortMemPerCore = int(self.params.samtoolsSortMemPerCore)
        minSortMem = 1000000
        if self.params.samtoolsSortMemPerCore < minSortMem :
            raise Exception("samtoolsSortMemPerCore must be an integer greater than minSortMem")

        if self.params.genomeFasta == None:
            raise Exception("No bwa genome file defined.")
        else:
            if not os.path.isfile(self.params.genomeFasta) :
                raise Exception("Can't find bwa genome file '%s'" % (self.params.genomeFasta))


    def workflow(self) :

        alignTasks = casava18BclToBamListFlow(self)
        mergeTask = self.addWorkflowTask("mergeBams", FlowcellBamListMergeFlow(self.params), dependencies=alignTasks)






#!/usr/bin/env python

import unittest
import os
import sys

scriptDir=os.path.abspath(os.path.dirname(__file__))

def pyflow_lib_dir() :
    return os.path.abspath(os.path.join(scriptDir,os.pardir,os.pardir,"pyflow","src"))

try :
    # if pyflow is in PYTHONPATH already then use the specified copy:
    from pyflow import isWindows,WorkflowRunner
except :
    # otherwise use the relative path within this repo:
    sys.path.append(pyflow_lib_dir())
    from pyflow import isWindows,WorkflowRunner


def getRmCmd() :
    if isWindows():
        return ["del","/f"]
    else:
        return ["rm","-f"]


def getSleepCmd() :
    if isWindows():
        return ["timeout"]
    else:
        return ["sleep"]


def getCatCmd() :
    if isWindows():
        return ["type"]
    else:
        return ["cat"]


def getCmdString(cmdList) :
    return " ".join(cmdList)



class NullWorkflow(WorkflowRunner) :
    pass



class TestWorkflowRunner(unittest.TestCase) :

    def __init__(self, *args, **kw) :
        unittest.TestCase.__init__(self, *args, **kw)
        self.testPath="testDataRoot"

    def setUp(self) :
        self.clearTestPath()

    def tearDown(self) :
        self.clearTestPath()

    def clearTestPath(self) :
        import shutil
        if os.path.isdir(self.testPath) :
            shutil.rmtree(self.testPath)


    def test_createDataDir(self) :
        w=NullWorkflow()
        w.run("local",self.testPath,isQuiet=True)
        self.assertTrue(os.path.isdir(self.testPath))


    def test_badMode(self) :
        w=NullWorkflow()
        try:
            w.run("foomode",self.testPath,isQuiet=True)
            self.fail("Didn't raise Exception")
        except KeyError:
            self.assertTrue(sys.exc_info()[1].args[0].find("foomode") != -1)


    def test_errorLogPositive(self) :
        """
        Test that errors are written to separate log when requested
        """
        os.mkdir(self.testPath)
        logFile=os.path.join(self.testPath,"error.log")
        w=NullWorkflow()
        try:
            w.run("foomode",self.testPath,errorLogFile=logFile,isQuiet=True)
            self.fail("Didn't raise Exception")
        except KeyError:
            self.assertTrue(sys.exc_info()[1].args[0].find("foomode") != -1)
        self.assertTrue((os.path.getsize(logFile) > 0))


    def test_errorLogNegative(self) :
        """
        Test that no errors are written to separate error log when none occur
        """
        os.mkdir(self.testPath)
        logFile=os.path.join(self.testPath,"error.log")
        w=NullWorkflow()
        w.run("local",self.testPath,errorLogFile=logFile,isQuiet=True)
        self.assertTrue((os.path.getsize(logFile) == 0))


    def test_dataDirCollision(self) :
        """
        Test that when two pyflow jobs are launched with the same dataDir, the second will fail.
        """
        import threading,time

        class StallWorkflow(WorkflowRunner) :
            def workflow(self2) :
                self2.addTask("sleeper",getSleepCmd()+["5"])

        class runner(threading.Thread) :
            def __init__(self2) :
                threading.Thread.__init__(self2)
                self2.retval1=1

            def run(self2) :
                w=StallWorkflow()
                self2.retval1=w.run("local",self.testPath,isQuiet=True)

        w2=StallWorkflow()

        r1=runner()
        r1.start()
        time.sleep(1)
        retval2=w2.run("local",self.testPath,isQuiet=True)
        self.assertTrue(retval2==1)
        r1.join()
        self.assertTrue(r1.retval1==0)


    def test_forceContinue(self) :
        class TestWorkflow(WorkflowRunner) :
            color="red"

            def setColor(self2,color) :
                self2.color=color

            def workflow(self2) :
                self2.addTask("A","echo "+self2.color)

        w=TestWorkflow()
        retval=w.run("local",self.testPath,isQuiet=True)
        self.assertTrue(retval==0)
        retval=w.run("local",self.testPath,isContinue=True,isQuiet=True)
        self.assertTrue(retval==0)
        w.setColor("green")
        retval=w.run("local",self.testPath,isContinue=True,isQuiet=True)
        self.assertTrue(retval==1)
        retval=w.run("local",self.testPath,isContinue=True,isForceContinue=True,isQuiet=True)
        self.assertTrue(retval==0)


    def test_badContinue(self) :
        w=NullWorkflow()
        try:
            w.run("local",self.testPath,isContinue=True,isQuiet=True)
            self.fail("Didn't raise Exception")
        except Exception:
            self.assertTrue(sys.exc_info()[1].args[0].find("Cannot continue run") != -1)


    def test_goodContinue(self) :
        w=NullWorkflow()
        retval1=w.run("local",self.testPath,isQuiet=True)
        retval2=w.run("local",self.testPath,isContinue=True,isQuiet=True)
        self.assertTrue((retval1==0) and (retval2==0))


    def test_autoContinue(self) :
        w=NullWorkflow()
        retval1=w.run("local",self.testPath,isContinue="Auto",isQuiet=True)
        retval2=w.run("local",self.testPath,isContinue="Auto",isQuiet=True)
        self.assertTrue((retval1==0) and (retval2==0))


    def test_simpleDependency(self) :
        "make sure B waits for A"
        class TestWorkflow(WorkflowRunner) :
            def workflow(self2) :
                filePath=os.path.join(self.testPath,"tmp.txt")
                self2.addTask("A","echo foo > " +filePath)
                self2.addTask("B",getCmdString(getCatCmd()) + " " + filePath + " && " + getCmdString(getRmCmd())+ " " + filePath,dependencies="A")

        w=TestWorkflow()
        self.assertTrue((0==w.run("local",self.testPath,isQuiet=True)))


    def test_waitDependency(self) :
        "make sure waitForTasks waits for A on the workflow thread"
        class TestWorkflow(WorkflowRunner) :
            def workflow(self2) :
                filePath=os.path.join(self.testPath,"tmp.txt")
                if os.path.isfile(filePath) : os.remove(filePath)
                self2.addTask("A",getCmdString(getSleepCmd()) + " 5 && echo foo > %s" % (filePath))
                self2.waitForTasks("A")
                assert(os.path.isfile(filePath))
                self2.addTask("B",getCmdString(getCatCmd()) + " " + filePath +" && " + getCmdString(getRmCmd())+ " " + filePath)

        w=TestWorkflow()
        self.assertTrue(0==w.run("local",self.testPath,isQuiet=True))


    def test_flowLog(self) :
        "make sure flowLog doesn't throw -- but this does not check if the log is updated"
        class TestWorkflow(WorkflowRunner) :
            def workflow(self2) :
                self2.flowLog("My Message")

        w=TestWorkflow()
        self.assertTrue(0==w.run("local",self.testPath,isQuiet=True))


    def test_deadSibling(self) :
        """
        Tests that when a task error occurs in one sub-workflow, its
        sibling workflows exit correctly (instead of hanging forever).
        This test is an early library error case.
        """
        class SubWorkflow1(WorkflowRunner) :
            "this one fails"
            def workflow(self2) :
                self2.addTask("A",getSleepCmd()+["5"])
                self2.addTask("B","boogyman!",dependencies="A")
                
        class SubWorkflow2(WorkflowRunner) :
            "this one doesn't fail"
            def workflow(self2) :
                self2.addTask("A",getSleepCmd()+["5"])
                self2.addTask("B",getSleepCmd()+["5"],dependencies="A")

        class MasterWorkflow(WorkflowRunner) :
            def workflow(self2) :
                wflow1=SubWorkflow1()
                wflow2=SubWorkflow2()
                self2.addWorkflowTask("wf1",wflow1)
                self2.addWorkflowTask("wf2",wflow2)

        w=MasterWorkflow()
        self.assertTrue(1==w.run("local",self.testPath,nCores=2,isQuiet=True))


    def test_selfDependency1(self) :
        """
        """
        class SelfWorkflow(WorkflowRunner) :
            def workflow(self2) :
                self2.addTask("A",getSleepCmd()+["5"],dependencies="A")
                
        w=SelfWorkflow()
        self.assertTrue(1==w.run("local",self.testPath,isQuiet=True))


    def test_expGraphScaling(self) :
        """
        This tests that pyflow does not scale poorly with highly connected subgraphs.

        When the error occurs, it locks the primary thread, so we put the test workflow
        on its own thread so that we can time it and issue an error.

        Issue reported by R Kelley and A Halpern
        """

        import threading

        class ScalingWorkflow(WorkflowRunner) :
            def workflow(self2) :
                tasks = set()
                for idx in xrange(60) :
                    sidx = str(idx)
                    tasks.add(self2.addTask("task_" + sidx, "echo " + sidx, dependencies = tasks))
                self2.waitForTasks("task_50")
                tasks.add(self2.addTask("task_1000", "echo 1000", dependencies = tasks))

        class runner(threading.Thread) :
            def __init__(self2) :
                threading.Thread.__init__(self2)
                self2.setDaemon(True)

            def run(self2) :
                w=ScalingWorkflow()
                w.run("local",self.testPath,isQuiet=True)

        r1=runner()
        r1.start()
        r1.join(30)
        self.assertTrue(not r1.isAlive())

    def test_startFromTasks(self) :
        """
        run() option to ignore all tasks before a specified task node
        """
        filePath=os.path.join(self.testPath,"tmp.txt")

        class SelfWorkflow(WorkflowRunner) :
            def workflow(self2) :
                self2.addTask("A","echo foo > "+filePath)
                self2.addTask("B",getSleepCmd()+["1"],dependencies="A")
                self2.addTask("C",getSleepCmd()+["1"],dependencies=("A","B"))
 
        w=SelfWorkflow()
        self.assertTrue(0==w.run("local",self.testPath,isQuiet=True,startFromTasks="B"))
        self.assertTrue(not os.path.exists(filePath))


    def test_startFromTasksSubWflow(self) :
        """
        run() option to ignore all tasks before a specified task node
        """
        filePath=os.path.join(self.testPath,"tmp.txt")

        class SubWorkflow(WorkflowRunner) :
            def workflow(self2) :
                self2.addTask("D","echo foo > "+filePath)

        class SelfWorkflow(WorkflowRunner) :
            def workflow(self2) :
                self2.addTask("A",getSleepCmd()+["1"])
                self2.addWorkflowTask("B",SubWorkflow(),dependencies="A")
                self2.addTask("C",getSleepCmd()+["1"],dependencies=("A","B"))

        w=SelfWorkflow()
        self.assertTrue(0==w.run("local",self.testPath,isQuiet=True,startFromTasks="B"))
        self.assertTrue(os.path.exists(filePath))


    def test_startFromTasksSubWflow2(self) :
        """
        run() option to ignore all tasks before a specified task node
        """
        filePath=os.path.join(self.testPath,"tmp.txt")

        class SubWorkflow(WorkflowRunner) :
            def workflow(self2) :
                self2.addTask("D","echo foo > "+filePath)

        class SelfWorkflow(WorkflowRunner) :
            def workflow(self2) :
                self2.addTask("A",getSleepCmd()+["1"])
                self2.addWorkflowTask("B",SubWorkflow(),dependencies="A")
                self2.addTask("C",getSleepCmd()+["1"],dependencies=("A","B"))

        w=SelfWorkflow()
        self.assertTrue(0==w.run("local",self.testPath,isQuiet=True,startFromTasks="C"))
        self.assertTrue(not os.path.exists(filePath))


    def test_ignoreTasksAfter(self) :
        """
        run() option to ignore all tasks below a specified task node
        """
        class SelfWorkflow(WorkflowRunner) :
            def workflow(self2) :
                self2.addTask("A",getSleepCmd()+["1"])
                self2.addTask("B",getSleepCmd()+["1"],dependencies="A")
                self2.addTask("C",getSleepCmd()+["1"],dependencies=("A","B"))
 
        w=SelfWorkflow()
        self.assertTrue(0==w.run("local",self.testPath,isQuiet=True,ignoreTasksAfter="B"))
        self.assertTrue(not w.isTaskComplete("C"))

    def test_addTaskOutsideWorkflow(self) :
        """
        test that calling addTask() outside of a workflow() method
        raises an exception
        """

        class SelfWorkflow(WorkflowRunner) :
            def __init__(self2) :
                self2.addTask("A",getSleepCmd()+["1"])

        try :
            w=SelfWorkflow()
            self.fail("Didn't raise Exception")
        except :
            pass

    def test_runModeInSubWorkflow(self) :
        """
        test that calling getRunMode() in a sub-workflow() method
        does not raise an exception (github issue #5)
        """

        class SubWorkflow(WorkflowRunner) :
            def workflow(self2) :
                if self2.getRunMode() == "local" :
                    self2.addTask("D",getSleepCmd()+["1"])

        class SelfWorkflow(WorkflowRunner) :
            def workflow(self2) :
                self2.addTask("A",getSleepCmd()+["1"])
                self2.addWorkflowTask("B",SubWorkflow(),dependencies="A")
                self2.addTask("C",getSleepCmd()+["1"],dependencies=("A","B"))

        try :
            w=SelfWorkflow()
            self.assertTrue(0==w.run("local",self.testPath,isQuiet=True))
        except :
            self.fail("Should not raise Exception")

    def test_CheckpointChain(self) :
        """
        Test that checkout points are handled correctly even
        when multiple checkpoints have a parent-child relationship
        """

        class SelfWorkflow(WorkflowRunner) :
            def workflow(self2) :
                self2.addTask("A")
                self2.addTask("B")
                self2.addTask("C",dependencies=["A","B"])

        try :
            w=SelfWorkflow()
            self.assertTrue(0==w.run("local",self.testPath,isQuiet=True))
        except :
            self.fail("Should not raise Exception")

if __name__ == '__main__' :
    unittest.main()




#!/usr/bin/env python
#
# pyFlow - a lightweight parallel task engine
#
# Copyright (c) 2012-2015 Illumina, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
# WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#


#
# demonstrate/test addTask() cwd option
#

import os.path
import sys

# add module path by hand
#
scriptDir=os.path.abspath(os.path.dirname(__file__))
sys.path.append(scriptDir+"/../../src")

from pyflow import WorkflowRunner


# all pyflow workflows are written into classes derived from
# pyflow.WorkflowRunner:
#
class CwdWorkflow(WorkflowRunner) :

    # a workflow is defined by overloading the
    # WorkflowRunner.workflow() method:
    #
    def workflow(self) :

        # get cwd and its parent for the addTask cwd test
        #
        cwd=os.getcwd()
        parentdir=os.path.abspath(os.path.join(cwd,".."))

        self.flowLog("testing pyflow cwd: '%s' parentdir: '%s'" % (cwd,parentdir))

        # task will fail unless pwd == parentdir:
        #
        # test both absolute and relative cwd arguments:
        #
        self.addTask("testAbsCwd","[ $(pwd) == '%s' ]; exit $?" % (parentdir),cwd=parentdir)
        self.addTask("testRelCwd","[ $(pwd) == '%s' ]; exit $?" % (parentdir),cwd="..")



# Instantiate the workflow
#
wflow = CwdWorkflow()

# Run the worklow:
#
retval=wflow.run(mode="local")

sys.exit(retval)




from distutils.core import setup

setup(
      name='pyFlow',
      version='${VERSION}',
      description='A lightweight parallel task engine',
      author='Chris Saunders',
      author_email='csaunders@illumina.com',
      packages=['pyflow'],
      package_dir={'pyflow': 'src'}
)



#!/usr/bin/env python
#
# pyFlow - a lightweight parallel task engine
#
# Copyright (c) 2012-2015 Illumina, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
# WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#


import os.path
import sys

# add module path by hand
#
scriptDir = os.path.abspath(os.path.dirname(__file__))
sys.path.append(scriptDir + "/../../src")

from pyflow import WorkflowRunner


#
# very simple task scripts called by the demo:
#
testJobDir = os.path.join(scriptDir, "testtasks")

sleepjob = os.path.join(testJobDir, "sleeper.bash")  # sleeps



# all pyflow workflows are written into classes derived from
# pyflow.WorkflowRunner:
#
class MutexWorkflow(WorkflowRunner) :

    # a workflow is defined by overloading the
    # WorkflowRunner.workflow() method:
    #
    def workflow(self) :

        # create an array of mutex restricted tasks which can only run
        # once at a time:
        for i in range(8) :
            self.addTask("mutex_task_" + str(i), sleepjob + " 1", mutex="test")

        # and add an array of 'normal' tasks for comparison:
        for i in range(16) :
            self.addTask("normal_task_" + str(i), sleepjob + " 1")




def main() :
    # Instantiate the workflow
    wflow = MutexWorkflow()

    # Run the worklow:
    retval = wflow.run(mode="local", nCores=6)

    sys.exit(retval)



if __name__ == "__main__" :
    main()



#!/usr/bin/env python
#
# pyFlow - a lightweight parallel task engine
#
# Copyright (c) 2012-2015 Illumina, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
# WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#


import os.path
import sys

# add module paths
#
filePath = os.path.dirname(__file__)
pyflowPath = os.path.abspath(os.path.join(filePath, "../../src"))
sys.path.append(pyflowPath)

from pyflow import WorkflowRunner
from getDemoRunOptions import getDemoRunOptions


#
# very simple task scripts called by the demo:
#
testJobDir = os.path.join(filePath, "testtasks")

sleepjob = os.path.join(testJobDir, "sleeper.bash")  # sleeps
yelljob = os.path.join(testJobDir, "yeller.bash")  # generates some i/o
runjob = os.path.join(testJobDir, "runner.bash")  # runs at 100% cpu



# all pyflow workflows are written into classes derived from
# pyflow.WorkflowRunner:
#
class TestWorkflow(WorkflowRunner) :

    # a workflow is defined by overloading the
    # WorkflowRunner.workflow() method:
    #
    def workflow(self) :

        # A simple command task with no dependencies, labeled 'task1'.
        #
        cmd = "%s 1" % (yelljob)
        self.addTask("task1", cmd)

        # Another task which runs the same command, this time the
        # command is provided as an argument list. An argument list
        # can be useful when a command has many arguments or
        # complicated quoting issues:
        #
        cmd = [yelljob, "1"]
        self.addTask("task2", cmd)

        # This task will always run on the local machine, no matter
        # what the run mode is. The force local option is useful for
        # non-cpu intensive jobs which are taking care of minor
        # workflow overhead (moving/touching files, etc)
        #
        self.addTask("task3a", sleepjob + " 10", isForceLocal=True)


# get runtime options
#
runOptions = getDemoRunOptions()

# Instantiate the workflow
#
wflow = TestWorkflow()

# Run the worklow with runtime options specified on the command-line:
#
retval = wflow.run(mode=runOptions.mode,
                 nCores=runOptions.jobs,
                 memMb=runOptions.memMb,
                 mailTo=runOptions.mailTo,
                 isContinue=(runOptions.isResume and "Auto" or False),
                 isForceContinue=True,
                 schedulerArgList=runOptions.schedulerArgList)

sys.exit(retval)




#!/usr/bin/env python
#
# pyFlow - a lightweight parallel task engine
#
# Copyright (c) 2012-2015 Illumina, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
# WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#


import os.path
import sys

# add module path by hand
#
scriptDir=os.path.abspath(os.path.dirname(__file__))
sys.path.append(scriptDir+"/../../src")

from pyflow import WorkflowRunner


#
# very simple task scripts called by the demo:
#
testJobDir=os.path.join(scriptDir,"testtasks")

sleepjob=os.path.join(testJobDir,"sleeper.bash")  # sleeps
yelljob=os.path.join(testJobDir,"yeller.bash")    # generates some i/o
runjob=os.path.join(testJobDir,"runner.bash")     # runs at 100% cpu



# all pyflow workflows are written into classes derived from
# pyflow.WorkflowRunner:
#
class SimpleWorkflow(WorkflowRunner) :

    # WorkflowRunner objects can create regular constructors to hold
    # run parameters or other state information:
    #
    def __init__(self,params) :
        self.params=params


    # a workflow is defined by overloading the
    # WorkflowRunner.workflow() method:
    #
    def workflow(self) :

        # A simple command task with no dependencies, labeled 'task1'.
        #
        cmd="%s 1" % (yelljob)
        self.addTask("task1",cmd)

        # Another task which runs the same command, this time the
        # command is provided as an argument list. An argument list
        # can be useful when a command has many arguments or
        # complicated quoting issues:
        #
        cmd=[yelljob,"1"]
        self.addTask("task2",cmd)

        # This task will always run on the local machine, no matter
        # what the run mode is. The force local option is useful for
        # non-cpu intensive jobs which are taking care of minor
        # workflow overhead (moving/touching files, etc)
        #
        self.addTask("task3a",sleepjob+" 10",isForceLocal=True)

        # This job is requesting 2 threads:
        #
        self.addTask("task3b",runjob+" 10",nCores=2)

        # This job is requesting 2 threads and 3 gigs of ram:
        #
        self.addTask("task3c",runjob+" 10",nCores=2,memMb=3*1024)


        # addTask and addWorkflowTask always return their task labels
        # as a simple convenience. taskName is set to "task4" now.
        #
        taskName=self.addTask("task4",sleepjob+" 1")

        # an example task dependency:
        #
        # pyflow stores dependencies in set() objects, but you can
        # provide a list,tuple,set or single string as the argument to
        # dependencies:
        #
        # all the task5* tasks below specify "task4" as their
        # dependency:
        #
        self.addTask("task5a",yelljob+" 2",dependencies=taskName)
        self.addTask("task5b",yelljob+" 2",dependencies="task4")
        self.addTask("task5c",yelljob+" 2",dependencies=["task4"])
        self.addTask("task5d",yelljob+" 2",dependencies=[taskName])

        # this time we launch a number of sleep tasks based on the
        # workflow parameters:
        #
        # we store all tasks in sleepTasks -- which we use to make
        # other tasks wait for this entire set of jobs to complete:
        #
        sleepTasks=set()
        for i in range(self.params["numSleepTasks"]) :
            taskName="sleep_task%i" % (i)
            sleepTasks.add(taskName)
            self.addTask(taskName,sleepjob+" 1",dependencies="task5a")

            ## note the three lines above could have been written in a
            ## more compact single-line format:
            ##
            #sleepTasks.add(self.addTask("sleep_task%i" % (i),sleepjob+" 1",dependencies="task5a"))

        # this job cannot start until all tasks in the above loop complete:
        self.addTask("task6",runjob+" 2",nCores=3,dependencies=sleepTasks)

        # This task is supposed to fail, uncomment to see error reporting:
        #
        #self.addTask("task7",sleepjob)

        # Note that no command is provided to this task. It will not
        # be distributed locally or to sge, but does provide a
        # convenient label for a set of tasks that other processes
        # depend on. There is no special "checkpoint-task" type in
        # pyflow -- but any task can function like one per this
        # example:
        #
        self.addTask("checkpoint_task",dependencies=["task1","task6","task5a"])

        # The final task depends on the above checkpoint:
        #
        self.addTask("task8",yelljob+" 2",dependencies="checkpoint_task")



# simulated workflow parameters
#
myRunParams={"numSleepTasks" : 15}


# Instantiate the workflow
#
# parameters are passed into the workflow via its constructor:
#
wflow = SimpleWorkflow(myRunParams)

# Run the worklow:
#
retval=wflow.run(mode="local",nCores=8)

sys.exit(retval)




#!/usr/bin/env python
#
# pyFlow - a lightweight parallel task engine
#
# Copyright (c) 2012-2015 Illumina, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
# WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#


#
# This demo shows possibly the simplist possible pyflow we can create --
# a single 'hello world' task. After experimenting with this file
# please see the 'simpleDemo' for coverage of a few more pyflow features
#

import os.path
import sys

# add module path
#
scriptDir=os.path.abspath(os.path.dirname(__file__))
sys.path.append(os.path.abspath(os.path.join(scriptDir,os.pardir,os.pardir,"src")))

from pyflow import WorkflowRunner


# all pyflow workflows are written into classes derived from pyflow.WorkflowRunner:
#
class HelloWorkflow(WorkflowRunner) :

    # a workflow is defined by overloading the WorkflowRunner.workflow() method:
    #
    def workflow(self) :

        #
        # The output for this task will be written to the file helloWorld.out.txt
        #
        self.addTask("easy_task1", "echo 'Hello World!' > helloWorld.out.txt")



# Instantiate the workflow
#
wflow = HelloWorkflow()

# Run the worklow:
#
retval = wflow.run()

# done!
sys.exit(retval)




#!/usr/bin/env python

#
# This demo shows possibly the simplist possible pyflow we can create -- 
# a single 'hello world' task. After experimenting with this file
# please see the 'simpleDemo' for coverage of a few more pyflow features
#

import os.path
import sys

# add module path by hand
#
sys.path.append(os.path.abspath(os.path.dirname(__file__))+"/../pyflow/src")


from pyflow import WorkflowRunner


# all pyflow workflows are written into classes derived from pyflow.WorkflowRunner:
#
class MemTestWorkflow(WorkflowRunner) :

    # a workflow is defined by overloading the WorkflowRunner.workflow() method:
    #
    def workflow(self) :

        # The following is our first and only task for this workflow:
        self.addTask("easy_task1","echo 'Hello World!'")
        self.addTask("easy_task2","echo 'Hello World!'")
        self.addTask("easy_task3","echo 'Hello World!'")
        self.addTask("easy_task4","echo 'Hello World!'",memMb=1)



# Instantiate the workflow
#
wflow = MemTestWorkflow()

# Run the worklow:
#
retval=wflow.run(nCores=8,memMb=2049)

# done!
sys.exit(retval)




#!/usr/bin/env python
#
# pyFlow - a lightweight parallel task engine
#
# Copyright (c) 2012-2015 Illumina, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
# WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

"""
This demonstrates a run of a prototype BCL to BWA BAM workflow
created as a production-scale proof of concept for pyflow.

The bwa workflow is written into the BWAWorkflow object. See
bwaworkflow.py for implementation details of this class.

Finally, make sure configuration settings in BWAWorkflowConfig
are appropriate before running.
"""

import os, sys

scriptDir = os.path.abspath(os.path.dirname(__file__))
scriptName = os.path.basename(__file__)


runScript1 = """#!/usr/bin/env python
# BWAWorkflow run script auto-generated by command: %s

import os.path, sys

scriptDir=os.path.abspath(os.path.dirname(__file__))
sys.path.append('%s')
from bwaworkflow import BWAWorkflow

class WorkflowOptions(object) :
""" % (" ".join(sys.argv), scriptDir)

runScript2 = """
def get_run_options() :
    from optparse import OptionParser
    import textwrap

    epilog=\"""Note this script can be re-run to continue the workflow run in case of interuption.
Also note that dryRun option has limited utility when task definition depends on upstream task results,
in which case the dry run will not cover the full 'live' run task set)\"""

    # no epilog in py 2.4! hack-in the feature instead:
    class MyOptionParser(OptionParser) :
        def __init__(self, *args, **kwargs):
            self.myepilog = None
            try:
                self.myepilog = kwargs.pop('epilog')
            except KeyError:
                pass
            OptionParser.__init__(self,*args, **kwargs)

        def print_help(self,*args,**kwargs) :
            OptionParser.print_help(self,*args, **kwargs)
            if self.myepilog != None :
                sys.stdout.write("%s\\n" % (textwrap.fill(self.myepilog)))

    parser = MyOptionParser(epilog=epilog)


    parser.add_option("-m", "--mode", type="string",dest="mode",
                      help="select run mode (local|sge)")
    parser.add_option("-j", "--jobs", type="string",dest="jobs",
	              help="number of jobs (default: 1 for local mode, 'unlimited' for sge mode)")
    parser.add_option("-e","--mailTo", type="string",dest="mailTo",action="append",
	              help="send email notification of job completion status to this address (may be provided multiple times for more than one email address)")
    parser.add_option("-d","--dryRun", dest="isDryRUn",action="store_true",
                      help="dryRun workflow code without actually running command-tasks")


    (options,args) = parser.parse_args()

    if len(args) :
        parser.print_help()
	sys.exit(2)

    if options.mode == None :
        parser.print_help()
	sys.exit(2)
    elif options.mode not in ["local","sge"] :
        parser.error("Invalid mode. Available modes are: local, sge")

    if options.jobs == None :
        if options.mode == "sge" :
	    options.jobs == "unlimited"
	else :
	    options.jobs == "1"
    elif (options.jobs != "unlimited") and (int(options.jobs) <= 0) :
        parser.error("Jobs must be 'unlimited' or an integer greater than 1")

    return options

runOptions=get_run_options()
flowOptions=WorkflowOptions()
flowOptions.outputDir=scriptDir
wflow = BWAWorkflow(flowOptions)
retval=wflow.run(mode=runOptions.mode,
                 nCores=runOptions.jobs,
                 dataDirRoot=scriptDir,
                 mailTo=runOptions.mailTo,
		 isContinue="Auto",
                 isForceContinue=True,
                 isDryRun=runOptions.isDryRUn)
sys.exit(retval)
"""




def checkArg(x, label, checkfunc) :
    if x != None:
        x = os.path.abspath(x)
        if not checkfunc(x) :
            raise Exception("Can't find %s: '%s'" % (label, x))
    return x

def checkDirArg(dir, label) :
    return checkArg(dir, label, os.path.isdir)

def checkFileArg(file, label) :
    return checkArg(file, label, os.path.isfile)



def get_option_parser(defaults, configFileName, isAllHelp=False) :
    from optparse import OptionGroup, OptionParser, SUPPRESS_HELP
    import textwrap

    description = """This script configures a bcl to BWA alignmed BAM workflow.
Given a bcl basecalls directory the workflow will create fastq's using CASAVA's
bcl to fastq converter, then align each fastq using bwa, and finally consolidate
the output into a single BAM file for for each Project/Sample combination.

The configuration process will produce a workflow run script, which can be used to
execute the workflow on a single node or through sge with a specific job limit.
"""

    epilog = """Default parameters will always be read from the file '%s' if it exists.
This file is searched for in the current working directory first -- if
it is not found then the directory containing this script is searched as well.
The current set of default parameters may be written to this file using the --writeConfig switch,
which takes all current defaults and arguments, writes these to the
configuration file and exits without setting up a workflow run script as usual.
""" % (configFileName)

    # no epilog in py 2.4! hack-in the feature instead:
    class MyOptionParser(OptionParser) :
        def __init__(self, *args, **kwargs):
            self.myepilog = None
            try:
                self.myepilog = kwargs.pop('epilog')
            except KeyError:
                pass
            OptionParser.__init__(self, *args, **kwargs)

        def print_help(self, *args, **kwargs) :
            OptionParser.print_help(self, *args, **kwargs)
            if self.myepilog != None :
                sys.stdout.write("%s\n" % (textwrap.fill(self.myepilog)))


    parser = MyOptionParser(description=description, epilog=epilog)

    parser.set_defaults(**defaults)

    parser.add_option("--allHelp", action="store_true", dest="isAllHelp",
                      help="show all extended/hidden options")

    group = OptionGroup(parser, "Workflow options")
    group.add_option("--bclBasecallsDir", type="string", dest="bclBasecallsDirList", metavar="DIR", action="append",
                      help="BCL basecalls directory. Call this option multiple times to specify multiple bcl directories, samples with the same name will be combined over all flowcells after alignmnet. [required] (default: %default)")
    group.add_option("--bclTilePattern", type="string", dest="bclTilePatternList", metavar="PATTERN", action="append",
                      help="BCL converter tiles expression used to select a subsset of tiles (eg. 's_1') call this option either once for each basecalls dir or not at all (default: %default)")
    group.add_option("--genomeFasta", type="string", dest="genomeFasta",
	              help="Genome fasta file which includes BWA index in the same directory [required] (default: %default)")
    group.add_option("--outputDir", type="string", dest="outputDir",
                      help="BCL basecalls directory [required] (default: %default)")
    group.add_option("--sampleName", type="string", dest="sampleNameList", metavar="sampleName", action="append",
                     help="Restrict analysis to given sampleName. This option can be provided more than once for multiple sample names. If no names are provided all samples are analyzed (default: %default)")
    parser.add_option_group(group)

    secgroup = OptionGroup(parser, "Extended options",
                                  "These options are not likely to be reset after initial configuration in a new site, they will not be printed here if a default exists from the configuration file or otherwise, unless --allHelp is specified")

    # used to access isAnyHelp from the maybeHelp function
    class Hack : isAnyHelp = False

    def maybeDefHelp(key, msg) :
        if isAllHelp or (key not in defaults) :
            Hack.isAnyHelp = True
            return msg
        return SUPPRESS_HELP

    secgroup.add_option("--casavaDir", type="string", dest="casavaDir",
                     help=maybeDefHelp("casavaDir", "casava 1.8.2+ installation directory [required] (default: %default)"))
    secgroup.add_option("--bwaBin", type="string", dest="bwaBin",
                     help=maybeDefHelp("bwaBin", "bwa binary [required] (default: %default)"))
    secgroup.add_option("--samtoolsBin", type="string", dest="samtoolsBin",
                     help=maybeDefHelp("samtoolsBin", "samtools binary [required] (default: %default)"))
    secgroup.add_option("--picardDir", type="string", dest="picardDir",
                     help=maybeDefHelp("picardDir", "casava 1.8.2+ installation directory [required] (default: %default)"))
    if not Hack.isAnyHelp:
        secgroup.description = "hidden"
    parser.add_option_group(secgroup)

    def maybeHelp(key, msg) :
        if isAllHelp : return msg
        return SUPPRESS_HELP

    configgroup = OptionGroup(parser, "Config options")
    configgroup.add_option("--writeConfig", action="store_true", dest="isWriteConfig",
                           help=maybeHelp("writeConfig", "Write new default configuration file based on current defaults and agruments. Defaults written to: '%s'" % (configFileName)))
    if not isAllHelp :
        configgroup.description = "hidden"
    parser.add_option_group(configgroup)

    return parser



def get_run_options() :
    from ConfigParser import SafeConfigParser

    configFileName = scriptName + ".ini"
    if not os.path.isfile(configFileName) :
        configPath = os.path.join(scriptDir, configFileName)
    else :
        configPath = os.path.join('.', configFileName)

    configSectionName = scriptName

    config = SafeConfigParser()
    config.optionxform = str
    config.read(configPath)

    configOptions = {}
    if config.has_section(configSectionName) :
        for (k, v) in config.items(configSectionName) :
            if v == "" : continue
            configOptions[k] = v

    defaults = { 'outputDir' : './results',
                 'bclToFastqMaxCores' : 12,
                 'samtoolsSortMemPerCore' : 1000000000,  # samtools sort uses about 2x what you tell it to...
                 'alnMaxCores' : 8,  # presumably bwa aln will become increasingly inefficient per core, so we don't want to let this go forever...
                 'isKeepFastq' : True,  # important to keep these during testing, but not for production
               }

    defaults.update(configOptions)

    parser = get_option_parser(defaults, configFileName)
    (options, args) = parser.parse_args()

    if options.isAllHelp :
        parser = get_option_parser(defaults, configFileName, True)
        parser.print_help()
        sys.exit(2)

    if len(args) :  # or (len(sys.argv) == 1):
        parser.print_help()
	sys.exit(2)

    # sanitize arguments before writing defaults, check for missing arguments after:
    #
    def checkListRepeats(list, itemLabel) :
        if list == None : return
        if len(set(list)) != len(list) :
            parser.error("Repeated %s entries" % (itemLabel))

    if options.bclBasecallsDirList != None :
        for i, bclDir in enumerate(options.bclBasecallsDirList) :
            options.bclBasecallsDirList[i] = checkDirArg(bclDir, "bcl basecalls directory")
    # tmp for testing:
    # checkListRepeats(options.bclBasecallsDirList,"bcl basecalls directory")
    if (options.bclTilePatternList != None) and \
       (len(options.bclBasecallsDirList) != len(options.bclTilePatternList)) :
        parser.error("Unexpected number of bclTilPattern entries")
    checkListRepeats(options.sampleNameList, "sample name")

    options.casavaDir = checkDirArg(options.casavaDir, "casava directory")

    options.genomeFasta = checkFileArg(options.genomeFasta, "genome fasta file")
    options.bwaBin = checkFileArg(options.bwaBin, "bwa binary")
    options.samtoolsBin = checkFileArg(options.samtoolsBin, "samtools binary")

    if options.isWriteConfig == True :
        if not config.has_section(configSectionName) :
            config.add_section(configSectionName)
        for k, v in vars(options).iteritems() :
            if k == "isWriteConfig" : continue
            if v == None : v = ""
            config.set(configSectionName, k, str(v))
        configfp = open(configFileName, "w")
        config.write(configfp)
        configfp.close()
        sys.exit(0)

    def noArgOrError(msg) :
        if len(sys.argv) <= 1 :
            parser.print_help()
            sys.exit(2)
        else :
            parser.error(msg)

    def assertOption(arg, label) :
        if arg == None:
            noArgOrError("No %s specified" % (label))

    def assertList(list, itemLabel) :
        if (list == None) or (len(list) == 0) :
            noArgOrError("List containing %s (s) is empty or missing" % (itemLabel))
        else :
            for item in list :
                assertOption(item, itemLabel)

    assertList(options.bclBasecallsDirList, "bcl basecalls directory")
    assertList(options.sampleNameList, "sample name")
    assertOption(options.genomeFasta, "genome fasta file")
    assertOption(options.outputDir, "output directory")
    assertOption(options.casavaDir, "casava directory")
    assertOption(options.picardDir, "picard directory")
    assertOption(options.bwaBin, "bwa binary")
    assertOption(options.samtoolsBin, "samtools binary")

    return options



from bwaworkflow import BWAWorkflow, ensureDir


def main() :

    options = get_run_options()

    # instantiate workflow object to trigger parameter validation only
    #
    wflow = BWAWorkflow(options)

    # generate runscript:
    #
    scriptFile = os.path.join(options.outputDir, "runWorkflow.py")
    ensureDir(options.outputDir)

    sfp = open(scriptFile, "w")
    sfp.write(runScript1)
    # there must be a nicer way to reverse eval() an object -- maybe human readable pickle is what we want here?
    for k, v in vars(options).iteritems() :
        if isinstance(v, basestring) :
            sfp.write("    %s = '%s'\n" % (k, v))
        else:
            sfp.write("    %s = %s\n" % (k, v))
    sfp.write("\n")
    sfp.write(runScript2)
    sfp.close()
    os.chmod(scriptFile, 0755)

    notefp = sys.stdout
    notefp.write("""
Successfully created workflow run script. To execute the workflow, run the following script and set appropriate options:

%s
""" % (scriptFile))


if __name__ == "__main__" :
    main()




#
# pyFlow - a lightweight parallel task engine
#
# Copyright (c) 2012-2015 Illumina, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
# WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

"""
pyflowConfig

This file defines a class instance 'siteConfig' containing pyflow components
which are the most likely to need site-specific configuration.
"""

import os


# this is the only object pyflow needs to import, it
# is defined at the end of this module:
#
siteConfig = None


# depending on network setup getfqdn() can be slow, so cache calls to this function here:
#
def _getHostName() :
    import socket
    return socket.getfqdn()

cachedHostName = None


def getHostName() :
    global cachedHostName
    if cachedHostName is None :
        cachedHostName = _getHostName()
    return cachedHostName


def getDomainName() :
    hn = getHostName().split(".")
    if len(hn) > 1 : hn = hn[1:]
    return ".".join(hn)



class DefaultSiteConfig(object) :
    """
    Default configuration settings are designed to work with as
    many sites as technically feasible
    """

    # All email notifications come from the following e-mail address
    #
    mailFrom = "pyflow-bot@" + getDomainName()

    # Default memory (in megabytes) requested by each command task:
    #
    defaultTaskMemMb = 2048

    # In local run mode, this is the defalt memory per thread that we
    # assume is available:
    #
    defaultHostMemMbPerCore = 2048

    # maximum number of jobs which can be submitted to sge at once:
    #
    # allowed values are "unlimited" or None for unlimited jobs, or
    # a positive integer describing the max number of jobs
    #
    maxSGEJobs = 128

    # both getHostName and getDomainName are used in the
    # siteConfig factory, so these are not designed to be
    # overridden at present:
    getHostName = staticmethod(getHostName)
    getDomainName = staticmethod(getDomainName)

    @classmethod
    def qsubResourceArg(cls, nCores, memMb) :
        """
        When a task is launched using qsub in sge mode, it will call this
        function to specify the requested number of threads and megabytes
        of memory. The returned argument list will be appended to the qsub
        arguments.

        nCores -- number of threads requested
        memMb -- memory requested (in megabytes)
        """
        nCores = int(nCores)
        memMb = int(memMb)
        return cls._qsubResourceArgConfig(nCores, memMb)

    @classmethod
    def _qsubResourceArgConfig(cls, nCores, memMb) :
        """
        The default function is designed for maximum
        portability  -- it just provides more memory
        via more threads.
        """

        # this is the memory we assume is available per
        # thread on the cluster:
        #
        class Constants(object) : megsPerCore = 4096

        memCores = 1 + ((memMb - 1) / Constants.megsPerCore)

        qsubCores = max(nCores, memCores)

        if qsubCores <= 1 : return []
        return ["-pe", "threaded", str(qsubCores)]


    @classmethod
    def getSgeMakePrefix(cls, nCores, memMb, schedulerArgList) :
        """
        This prefix will be added to ' -C directory', and run from
        a local process to handle sge make jobs.

        Note that memMb hasn't been well defined for make jobs yet,
        is it the per task memory limit? The first application to
        accually make use of this will have to setup the convention,
        it is ignored right now...
        """
        nCores = int(nCores)
        memMb = int(memMb)

        retval = ["qmake",
                "-V",
                "-now", "n",
                "-cwd",
                "-N", "pyflowMakeTask"]

        # user arguments to run() (usually q specification:
        retval.extend(schedulerArgList)

        #### use qmake parallel environment:
        # retval.extend(["-pe","make",str(nCores),"--"])

        #### ...OR use 'dynamic' sge make environment:
        retval.extend(["--", "-j", str(nCores)])

        return retval



def getEnvVar(key) :
    if key in os.environ : return os.environ[key]
    return None



class hvmemSGEConfig(DefaultSiteConfig) :
    """
    This config assumes 'h_vmem' is defined on the SGE instance

    """

    @classmethod
    def _qsubResourceArgConfig(cls, nCores, memMb) :

        retval = []

        # specify memory requirements
        memGb = 1 + ((memMb - 1) / 1024)
        reqArg = "h_vmem=%iG" % (memGb)
        retval.extend(["-l", reqArg])

        if nCores > 1 :
            retval.extend(["-pe", "threaded", str(nCores)])

        return retval



#
# final step is the selection of this run's siteConfig object:
#

def siteConfigFactory() :
    # TODO: add an import time test to determine if h_vmem is valid
    return  hvmemSGEConfig


siteConfig = siteConfigFactory()





#!/usr/bin/env python
#
# pyFlow - a lightweight parallel task engine
#
# Copyright (c) 2012-2015 Illumina, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
# WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#


import os.path
import sys

pyflowDir = os.path.abspath(os.path.join(os.path.dirname(__file__), "../../src"))
sys.path.append(pyflowDir)

from optparse import OptionParser, SUPPRESS_HELP
from pyflow import WorkflowRunner

from pyflow import isLocalSmtp


localDefaultCores = WorkflowRunner.runModeDefaultCores('local')
sgeDefaultCores = WorkflowRunner.runModeDefaultCores('sge')



def getDemoRunOptions() :
    """
    This routine is shared by a demo programs to demostrate how to pass pyflow's runtime options on to command-line options. It is not intended to be a demo program itself.
    """

    parser = OptionParser()


    parser.add_option("-m", "--mode", type="string", dest="mode",
                      help="Select run mode {local,sge} (required)")
    parser.add_option("-q", "--queue", type="string", dest="queue",
                      help="Specify sge queue name. Argument ignored if mode is not sge")
    parser.add_option("-j", "--jobs", type="string", dest="jobs",
	                  help="Number of jobs, must be an integer or 'unlimited' (default: %s for local mode, %s for sge mode)" % (localDefaultCores, sgeDefaultCores))
    parser.add_option("-g", "--memGb", type="string", dest="memGb",
	               help="Gigabytes of memory available to run workflow -- only meaningful in local mode, must be an integer or 'unlimited' (default: 2*jobs for local mode, 'unlimited' for sge mode)")
    parser.add_option("-r", "--resume", dest="isResume", action="store_true", default=False,
                      help="Resume a workflow from the point of interuption. This flag has no effect on a new workflow run.")

    isEmail = isLocalSmtp()
    emailHelp=SUPPRESS_HELP
    if isEmail:
        emailHelp="Send email notification of job completion status to this address (may be provided multiple times for more than one email address)"

    parser.add_option("-e", "--mailTo", type="string", dest="mailTo", action="append",
                      help=emailHelp)


    (options, args) = parser.parse_args()

    if not isEmail :
        options.mailTo = None

    if len(args) :
        parser.print_help()
        sys.exit(2)

    if options.mode is None :
        parser.print_help()
        sys.stderr.write("\n\nERROR: must specify run mode\n\n")
        sys.exit(2)
    elif options.mode not in ["local", "sge"] :
        parser.error("Invalid mode. Available modes are: local, sge")

    if options.jobs is None :
        if options.mode == "sge" :
            options.jobs = sgeDefaultCores
        else :
            options.jobs = localDefaultCores
    if options.jobs != "unlimited" :
        options.jobs = int(options.jobs)
        if options.jobs <= 0 :
            parser.error("Jobs must be 'unlimited' or an integer greater than 1")

    # note that the user sees gigs, but we set megs
    if options.memGb is None :
        if options.mode == "sge" :
            options.memMb = "unlimited"
        else :
            if options.jobs == "unlimited" :
                options.memMb = "unlimited"
            else :
                options.memMb = 2 * 1024 * options.jobs
    elif options.memGb != "unlimited" :
        options.memGb = int(options.memGb)
        if options.memGb <= 0 :
            parser.error("memGb must be 'unlimited' or an integer greater than 1")
        options.memMb = 1024 * options.memGb
    else :
        options.memMb = options.memGb

    options.schedulerArgList = []
    if options.queue is not None :
        options.schedulerArgList = ["-q", options.queue]

    return options



if __name__ == "__main__" :
    help(getDemoRunOptions)




#!/usr/bin/env python 
#
"""
automation friendly cross-platform tests for pyflow
"""

import os
import sys

scriptDir=os.path.abspath(os.path.dirname(__file__))


def getOptions() :

    from optparse import OptionParser

    usage = "usage: %prog [options]"
    parser = OptionParser(usage=usage,description="Run all pyflow tests")

    parser.add_option("--nosge",dest="isSkipSge", action="store_true",
                      help="skip SGE testing")

    (options,args) = parser.parse_args()

    if len(args) != 0 :
        parser.print_help()
        sys.exit(2)

    return (options,args)


def main() :
    import subprocess

    (options,args) = getOptions()

    pyflowRootDir=os.path.abspath(os.path.join(scriptDir,os.pardir,os.pardir))
    pyflowDir=os.path.join(pyflowRootDir,"pyflow")

    utScriptPath=os.path.join(scriptDir,"pyflow_unit_tests.py")

    if True :
        # process-out to run the unit tests for now -- TODO: can we just import this instead?
        utCmd=[sys.executable,"-E",utScriptPath,"-v"]
        proc = subprocess.Popen(utCmd)
        proc.wait()
        if proc.returncode != 0 :
            raise Exception("Pyflow unit test run failed") 

    # run through demos (only helloWorld is working on windows)
    if True :
        demoDir=os.path.join(pyflowDir,"demo")
        for demoName in ["helloWorld"] :
            demoScriptPath=os.path.join(demoDir,demoName,demoName+".py")
            demoCmd=[sys.executable,"-E",demoScriptPath]
            proc = subprocess.Popen(demoCmd)
            proc.wait()
            if proc.returncode != 0 :
                raise Exception("Pyflow demo failed: '%s'" % (demoScriptPath))


main()




#!/usr/bin/env python
#
# pyFlow - a lightweight parallel task engine
#
# Copyright (c) 2012-2015 Illumina, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
# WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#


#
# This is a very simple demo/test of pyFlow's new (@ v0.4) memory
# resource feature.
#

import os.path
import sys

# add module path by hand
#
sys.path.append(os.path.abspath(os.path.dirname(__file__)) + "/../../src")


from pyflow import WorkflowRunner


# all pyflow workflows are written into classes derived from pyflow.WorkflowRunner:
#
class MemTestWorkflow(WorkflowRunner) :

    # a workflow is defined by overloading the WorkflowRunner.workflow() method:
    #
    def workflow(self) :

        # Each task has a default memory request of 2048 megabytes
        # but this is site-configurable in pyflowConfig.py, so we
        # specify it for every task here
        #
        # This works correctly if task 4 is the only task run in
        # parallel with one of the other 3 tasks.
        #
        self.addTask("task1", "echo 'Hello World!'", memMb=2048)
        self.addTask("task2", "echo 'Hello World!'", memMb=2048)
        self.addTask("task3", "echo 'Hello World!'", memMb=2048)
        self.addTask("task4", "echo 'Hello World!'", memMb=1)



# Instantiate the workflow
#
wflow = MemTestWorkflow()

# Run the worklow:
#
retval = wflow.run(nCores=8, memMb=2049)

# done!
sys.exit(retval)




#!/usr/bin/env python
#
# pyFlow - a lightweight parallel task engine
#
# Copyright (c) 2012-2015 Illumina, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
# WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#


#
# demonstrate/test addTask() env option
#

import os.path
import sys

# add module path by hand
#
scriptDir = os.path.abspath(os.path.dirname(__file__))
sys.path.append(scriptDir + "/../../src")

from pyflow import WorkflowRunner


# all pyflow workflows are written into classes derived from
# pyflow.WorkflowRunner:
#
class EnvWorkflow(WorkflowRunner) :

    # a workflow is defined by overloading the
    # WorkflowRunner.workflow() method:
    #
    def workflow(self) :
        # run a task with the parent env:
        #
        home = os.environ["HOME"]
        self.addTask("testDefEnv", "[ $HOME == '%s' ]; exit $?" % (home))

        # create a minimal test environment
        #
        new_path = "/bin"
        min_env = { "PATH" : new_path }
        self.addTask("testMinEnv", "[ $PATH == '%s' ]; exit $?" % (new_path), env=min_env)

        # augment parent env with additional settings:
        #
        augmented_env = os.environ.copy()
        augmented_env["FOO"] = "BAZ"
        self.addTask("testAugmentedEnv", "[ $FOO == 'BAZ' ]; exit $?", env=augmented_env)

        # test funny characters that have shown to cause trouble on some sge installations
        funky_env = {}
        funky_env["PATH"] = "/bin"
        funky_env["_"] = "| %s %F \n"
        # in this case we just want the job to run at all:
        self.addTask("testFunkyEnv", "echo 'foo'; exit $?", env=funky_env)

        assert("FOO" not in os.environ)



# Instantiate the workflow
#
wflow = EnvWorkflow()

# Run the worklow:
#
retval = wflow.run(mode="local")

sys.exit(retval)




#!/usr/bin/env python

import os.path
import sys

sys.path.append(os.path.abspath(os.path.dirname(__file__)) + "/../../src")

import pyflow

# Document the public functions of pyflow's only public class:
#
help(pyflow.WorkflowRunner)




#!/usr/bin/env python
#
# pyFlow - a lightweight parallel task engine
#
# Copyright (c) 2012-2015 Illumina, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
# WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#


#
# This is a very simple demo/test of pyFlow's new (@ v0.4) memory
# resource feature.
#

import os.path
import sys

# add module path by hand
#
sys.path.append(os.path.abspath(os.path.dirname(__file__)) + "/../../src")

from pyflow import WorkflowRunner


# all pyflow workflows are written into classes derived from pyflow.WorkflowRunner:
#
class RetryWorkflow(WorkflowRunner) :

    # a workflow is defined by overloading the WorkflowRunner.workflow() method:
    #
    def workflow(self) :

        # this task behaves correctly it retries the job 4 times before failing, no automated way
        # to confirm success right now.
        #
        self.flowLog("****** NOTE: This demo is supposed to fail ******")
        self.addTask("retry_task_success", "exit 0", retryMax=8, retryWait=2, retryWindow=0, retryMode="all")
        self.addTask("retry_task_fail", "exit 1", retryMax=3, retryWait=2, retryWindow=0, retryMode="all")



# Instantiate the workflow
#
wflow = RetryWorkflow()

# Run the worklow:
#
retval = wflow.run()

if retval == 0 :
    raise Exception("Example workflow is expected to fail, but did not.")
else :
    sys.stderr.write("INFO: Demo workflow failed as expected.\n\n")


# Run the workflow again to demonstrate that global settings are overridden by task retry settings:
#
retval = wflow.run(retryMax=0)

if retval == 0 :
    raise Exception("Example workflow is expected to fail, but did not.")
else :
    sys.stderr.write("INFO: Demo workflow failed as expected.\n\n")





#!/usr/bin/env python

import os,sys,time
import datetime

if len(sys.argv) != 2 :
    print "usage $0 arg"
    sys.exit(1)

arg=sys.argv[1]

pid=os.getpid()

sys.stdout.write("pid: %s arg: %s starting yell\n" % (str(pid),arg))

for i in xrange(100):
    td=datetime.datetime.utcnow().isoformat()
    msg="Yeller %s yellin %i" % (str(pid),i)
    sys.stdout.write(msg+" stdout "+td+"\n")
    sys.stderr.write(msg+" stderr "+td+"\n")
    time.sleep(1)

sys.stdout.write("pid: %s arg: %s ending yell\n" % (str(pid),arg))




#
# pyFlow - a lightweight parallel task engine
#
# Copyright (c) 2012-2015 Illumina, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
# WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

"""
This script wraps workflow tasks for execution on local or remote
hosts.  It is responsible for adding log decorations to task's stderr
output (which is diverted to a file), and writing task state transition
and error information to the wrapper's stderr, which becomes the
task's 'signal' file from pyflow's perspective. The signal file is
used to determine task exit status, total runtime, and queue->run
state transition when pyflow is run in SGE mode.
"""

import datetime
import os
import subprocess
import sys
import time


scriptName = "pyflowTaskWrapper.py"


def getTracebackStr() :
    import traceback
    return traceback.format_exc()


def getExceptionMsg() :
    return ("[ERROR] Unhandled Exception in pyflowTaskWrapper\n" + getTracebackStr())


def timeStampToTimeStr(ts) :
    """
    converts timeStamp (time.time()) output to timeStr
    """
    return datetime.datetime.utcfromtimestamp(ts).isoformat()

def timeStrNow():
    return timeStampToTimeStr(time.time())

def hardFlush(ofp):
    ofp.flush()
    if ofp.isatty() : return
    os.fsync(ofp.fileno())

def isWindows() :
    import platform
    return (platform.system().find("Windows") > -1)

class SimpleFifo(object) :
    """
    Store up to last N objects, not thread safe.
    Note extraction does not follow any traditional fifo interface
    """

    def __init__(self, size) :
        self._size = int(size)
        assert (self._size > 0)
        self._data = [None] * self._size
        self._head = 0
        self._occup = 0
        self._counter = 0


    def count(self) :
        """
        Get the total number of adds for this fifo
        """
        return self._counter


    def add(self, obj) :
        """
        add obj to fifo, and return obj for convenience
        """
        self._data[self._head] = obj
        self._counter += 1
        if self._occup < self._size : self._occup += 1
        self._head += 1
        if self._head == self._size : self._head = 0
        assert (self._head < self._size)
        return obj


    def get(self) :
        """
        return an array of the fifo contents
        """
        retval = []
        current = (self._head + self._size) - self._occup
        for _ in range(self._occup) :
            while current >= self._size : current -= self._size
            retval.append(self._data[current])
            current += 1
        return retval



class StringBling(object) :
    def __init__(self, runid, taskStr) :
        def getHostName() :
            import socket
            # return socket.gethostbyaddr(socket.gethostname())[0]
            return socket.getfqdn()

        self.runid = runid
        self.taskStr = taskStr
        self.hostname = getHostName()

    def _writeMsg(self, ofp, msg, taskStr, writeFilter=lambda x: x) :
        """
        log a possibly multi-line message with decoration:
        """
        prefix = "[%s] [%s] [%s] [%s] " % (timeStrNow(), self.hostname, self.runid, taskStr)
        if msg[-1] == "\n" : msg = msg[:-1]
        for line in msg.split("\n") :
            ofp.write(writeFilter(prefix + line + "\n"))
        hardFlush(ofp)


    def transfer(self, inos, outos, writeFilter=lambda x: x):
        """
        This function is used to decorate the stderr stream from the launched task itself
        """
        #
        # write line-read loop this way to workaround python bug:
        # http://bugs.python.org/issue3907
        #
        while True:
            line = inos.readline()
            if not line: break
            self._writeMsg(outos, line, self.taskStr, writeFilter)

    def wrapperLog(self, log_os, msg) :
        """
        Used by the wrapper to decorate each msg line with a prefix. The decoration
        is similar to that for the task's own stderr, but we prefix the task with
        'pyflowTaskWrapper' to differentiate the source.
        """
        self._writeMsg(log_os, msg, "pyflowTaskWrapper:" + self.taskStr)



def getParams(paramsFile) :
    import pickle

    paramhash = pickle.load(open(paramsFile))
    class Params : pass
    params = Params()
    for (k, v) in paramhash.items() : setattr(params, k, v)
    return params



def main():

    usage = """

Usage: %s runid taskid parameter_pickle_file

The parameter pickle file contains all of the task parameters required by the wrapper

""" % (scriptName)

    def badUsage(msg=None) :
        sys.stderr.write(usage)
        if msg is not None :
            sys.stderr.write(msg)
            exitval = 1
        else:
            exitval = 2
        hardFlush(sys.stderr)
        sys.exit(exitval)

    def checkExpectArgCount(expectArgCount) :
        if len(sys.argv) == expectArgCount : return
        badUsage("Incorrect argument count, expected: %i observed: %i\n" % (expectArgCount, len(sys.argv)))


    runid = "unknown"
    taskStr = "unknown"

    if len(sys.argv) > 2 :
        runid = sys.argv[1]
        taskStr = sys.argv[2]

    bling = StringBling(runid, taskStr)

    # send a signal for wrapper start as early as possible to help ensure hostname is logged
    pffp = sys.stderr
    bling.wrapperLog(pffp, "[wrapperSignal] wrapperStart")

    checkExpectArgCount(4)

    picklefile = sys.argv[3]

    # try multiple times to read the argument file in case of NFS delay:
    #
    retryDelaySec = 30
    maxTrials = 3
    for _ in range(maxTrials) :
        if os.path.exists(picklefile) : break
        time.sleep(retryDelaySec)

    if not os.path.exists(picklefile) :
        badUsage("First argument does not exist: " + picklefile)

    if not os.path.isfile(picklefile) :
        badUsage("First argument is not a file: " + picklefile)

    # add another multi-trial loop on the pickle load operation --
    # on some filesystems the file can appear to exist but not
    # be fully instantiated yet:
    #
    for t in range(maxTrials) :
        try :
            params = getParams(picklefile)
        except :
            if (t+1) == maxTrials :
                raise
            time.sleep(retryDelaySec)
            continue
        break

    if params.cmd is None :
        badUsage("Invalid TaskWrapper input: task command set to NONE")

    if params.cwd == "" or params.cwd == "None" :
        params.cwd = None

    toutFp = open(params.outFile, "a")
    terrFp = open(params.errFile, "a")

    # always keep last N lines of task stderr:
    fifo = SimpleFifo(20)

    isWin=isWindows()

    # Present shell as arg list with Popen(shell=False), so that
    # we minimize quoting/escaping issues for 'cmd' itself:
    #
    fullcmd = []
    if (not isWin) and params.isShellCmd :
        # TODO shell selection should be configurable somewhere:
        shell = ["/bin/bash", "--noprofile", "-o", "pipefail"]
        fullcmd = shell + ["-c", params.cmd]
    else :
        fullcmd = params.cmd

    retval = 1

    isShell=isWin

    try:
        startTime = time.time()
        bling.wrapperLog(pffp, "[wrapperSignal] taskStart")
        # turn off buffering so that stderr is updated correctly and its timestamps
        # are more accurate:
        # TODO: is there a way to do this for stderr only?
        proc = subprocess.Popen(fullcmd, stdout=toutFp, stderr=subprocess.PIPE, shell=isShell, bufsize=1, cwd=params.cwd, env=params.env)
        bling.transfer(proc.stderr, terrFp, fifo.add)
        retval = proc.wait()

        elapsed = (time.time() - startTime)

        # communication back to pyflow:
        bling.wrapperLog(pffp, "[wrapperSignal] taskExitCode %i" % (retval))

        # communication to human-readable log:
        msg = "Task: '%s' exit code: '%i'" % (taskStr, retval)
        bling.wrapperLog(terrFp, msg)

        if retval == 0 :
            # communication back to pyflow:
            bling.wrapperLog(pffp, "[wrapperSignal] taskElapsedSec %i" % (int(elapsed)))

            # communication to human-readable log:
            msg = "Task: '%s' complete." % (taskStr)
            msg += " elapsedSec: %i" % (int(elapsed))
            msg += " elapsedCoreSec: %i" % (int(elapsed * params.nCores))
            msg += "\n"
            bling.wrapperLog(terrFp, msg)
        else :
            # communication back to pyflow:
            tailMsg = fifo.get()
            bling.wrapperLog(pffp, "[wrapperSignal] taskStderrTail %i" % (1 + len(tailMsg)))
            pffp.write("Last %i stderr lines from task (of %i total lines):\n" % (len(tailMsg), fifo.count()))
            for line in tailMsg :
                pffp.write(line)
            hardFlush(pffp)


    except KeyboardInterrupt:
        msg = "[ERROR] Keyboard Interupt, shutting down task."
        bling.wrapperLog(terrFp, msg)
        sys.exit(1)
    except:
        msg = getExceptionMsg()
        bling.wrapperLog(terrFp, msg)
        raise

    sys.exit(retval)



if __name__ == "__main__" :
    main()




#!/usr/bin/env python
#
# pyFlow - a lightweight parallel task engine
#
# Copyright (c) 2012-2015 Illumina, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
# WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#


import os.path
import sys

# add module path by hand
#
scriptDir=os.path.abspath(os.path.dirname(__file__))
sys.path.append(scriptDir+"/../../src")

from pyflow import WorkflowRunner


#
# test and demostrate the use of a custom success message
# at the end of a workflow
#



# all pyflow workflows are written into classes derived from
# pyflow.WorkflowRunner:
#
class SuccessWorkflow(WorkflowRunner) :

    # a workflow is defined by overloading the
    # WorkflowRunner.workflow() method:
    #
    def workflow(self) :

        # provide a minimum task
        self.addTask("task1","touch success\! && exit 0")




# Instantiate the workflow
#
wflow = SuccessWorkflow()

# Run the worklow:
#
cwd=os.getcwd()
successMsg  = "SuccessWorkflow has successfully succeeded!\n"
successMsg += "\tPlease find your token of successful succeeding here: '%s'\n" % (cwd)
retval=wflow.run(mode="local",nCores=8,successMsg=successMsg,mailTo="csaunders@illumina.com")

sys.exit(retval)




#!/usr/bin/env python
"""
this is a script which runs a lot of features...
it does not provide test coverage as to whether everything
ran correctly... it will only pick up a basic crash or hang.
"""

import os.path
import sys

# bad example of how to add the path:
scriptDir=os.path.abspath(os.path.dirname(__file__))


# setup PYTHONPATH instead...

#sys.path.append(scriptDir+"/../pyflow/src")

from pyflow import WorkflowRunner


testJobDir=os.path.join(scriptDir,"testtasks")

sleepjob=os.path.join(testJobDir,"sleeper.bash")
yelljob=os.path.join(testJobDir,"yeller.bash")
runjob=os.path.join(testJobDir,"runner.bash")

class SubSubWorkflow(WorkflowRunner) :

    def workflow(self) :
        self.addTask("easy_task1",yelljob+" 1")
        self.addTask("easy_task2",runjob+" 2",nCores=3,dependencies=["easy_task1"])
        self.waitForTasks("easy_task2")
        self.addTask("easy_task3",runjob+" 2",nCores=3,dependencies=["easy_task2"])
        # intentional fail:
        #self.addTask("easy_task3b",runjob,dependencies=["easy_task2"])


class SubWorkflow(WorkflowRunner) :

    def workflow(self) :
        self.addTask("easy_task1",yelljob+" 1")
        self.addTask("easy_task2",runjob+" 2",nCores=3,dependencies=["easy_task1"])
        self.addTask("easy_task3",runjob+" 2",nCores=3,dependencies=["easy_task2"])
        wflow=SubSubWorkflow()
        self.addWorkflowTask("subsubwf_task1",wflow,dependencies="easy_task1")



class TestWorkflow(WorkflowRunner) :

    def workflow(self) :

	job=sleepjob+" 1"

        self.addTask("easy_task1",yelljob+" 1")
        waitTask=self.addTask("easy_task3",runjob+" 10",nCores=2,memMb=1024,isForceLocal=True)
        self.flowLog("My message")

        swflow=SubWorkflow()

        self.addWorkflowTask("subwf_task1",swflow,dependencies=waitTask)
        self.addWorkflowTask("subwf_task2",swflow,dependencies=waitTask)

        self.addTask("easy_task4",runjob+" 2",nCores=3,dependencies=["subwf_task1","subwf_task2"])
        self.addTask("easy_task5",job,nCores=1)

        # and stop here
        self.waitForTasks()

        self.flowLog("ITC1: "+str(self.isTaskComplete("easy_task1")))
        self.flowLog("ITC6: "+str(self.isTaskComplete("easy_task6")))

        self.addTask("easy_task6",job)
        #self.addTask("easy_task2",sleepjob)
        self.addTask("checkpoint_task",dependencies=["easy_task1","easy_task6","easy_task4"])
        self.addTask("dep_task",sleepjob+" 4",dependencies=["checkpoint_task"])



def getRunOptions() :

    from optparse import OptionParser

    defaults = { "mode" : "local" }

    parser = OptionParser()
    parser.set_defaults(**defaults)

    parser.add_option("-m", "--mode", type="string", dest="mode",
                      help="Select run mode {local,sge} (default: %default)")

    (options, args) = parser.parse_args()

    if len(args) :
        parser.print_help()
        sys.exit(2)

    if options.mode not in ["sge","local"] :
        parser.print_help()
        sys.exit(2)

    return options



def main() :
    options = getRunOptions()
    wflow = TestWorkflow()
    retval=wflow.run(options.mode,nCores=8,memMb=8*1024,isContinue=False)
    sys.exit(retval)



if __name__ == "__main__" :
    main()



#!/usr/bin/env python
#
# pyFlow - a lightweight parallel task engine
#
# Copyright (c) 2012-2015 Illumina, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
# WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#


import os.path
import sys

# add module path by hand
#
scriptDir=os.path.abspath(os.path.dirname(__file__))
sys.path.append(scriptDir+"/../../src")

from pyflow import WorkflowRunner


#
# very simple task scripts called by the demo:
#
testJobDir=os.path.join(scriptDir,"testtasks")

sleepjob=os.path.join(testJobDir,"sleeper.bash")  # sleeps
yelljob=os.path.join(testJobDir,"yeller.bash")    # generates some i/o
runjob=os.path.join(testJobDir,"runner.bash")     # runs at 100% cpu



# all pyflow workflows are written into classes derived from pyflow.WorkflowRunner:
#
# this workflow  is a simple example of a workflow we can either run directly,
# or run as a task within another workflow:
#
class SubWorkflow(WorkflowRunner) :

    # a workflow is defined by overloading the WorkflowRunner.workflow() method:
    #
    def workflow(self) :
        # this workflow executes a simple dependency diamond:
        self.addTask("task1",yelljob+" 1")
        self.addTask("task2a",yelljob+" 1",dependencies="task1")
        self.addTask("task2b",yelljob+" 1",dependencies="task1")
        self.addTask("task3",yelljob+" 1",dependencies=("task2a","task2b"))


#
# This workflow will use SubWorkflow as a task:
#
class SimpleWorkflow(WorkflowRunner) :

    # a workflow is defined by overloading the WorkflowRunner.workflow() method:
    #
    def workflow(self) :

        # it's fine to repeat task names in two workflows, even if you're sub-tasking one from the other
        self.addTask("task1",yelljob+" 1")
        self.addTask("task2",runjob+" 3")

        # instantiate a new workflow and run it as soon as task1 and task2 complete
        wflow=SubWorkflow()
        self.addWorkflowTask("subwf_task3",wflow,dependencies=("task1","task2"))

        # this job will not run until the workflow-task completes. This means that all of the
        # tasks that SubWorkflow launches will need to complete successfully beforehand:
        #
        self.addTask("task4",sleepjob+" 1",dependencies="subwf_task3")


# Instantiate our workflow
#
wflow = SimpleWorkflow()

# Run the worklow:
#
retval=wflow.run(mode="local",nCores=8)


# If we want to run the SubWorkflow as a regular workflow, that can be done as well:
#

#wflow2 = SubWorkflow()
#retval2=wflow2.run()


sys.exit(retval)





#!/usr/bin/env python

import os.path
import sys

# add module path by hand
#
scriptDir=os.path.abspath(os.path.dirname(__file__))
sys.path.append(scriptDir+"/../../pyflow/src")

from pyflow import WorkflowRunner



class SimpleWorkflow(WorkflowRunner) :
    """
    A workflow designed to differentiate the runtime impact
    of STREL-391
    """

    def __init__(self) :
        pass

    def workflow(self) :
        for i in range(4000) :
            self.addTask("task%s" % (i),["sleep","0"])



# Instantiate the workflow
#
# parameters are passed into the workflow via its constructor:
#
wflow = SimpleWorkflow()

# Run the worklow:
#
retval=wflow.run(mode="local",nCores=400,isQuiet=True)

sys.exit(retval)

